unit rjMime;
{
Unit:           rjMime
Version:        1.20
Last Modified:  20. June 2000
Author:         Ralf Junker <ralfjunker@gmx.de>

Description:    Ligtening fast Mime (Base64) Encoding and Decoding routines.
                More detailed descriptions follow the declarations of the
                functions and procedures below.

Legal:          This software is provided 'as-is', without any express or
                implied warranty. In no event will the author be held liable
                for any  damages arising from the use of this software.

                Permission is granted to anyone to use this software for any
                purpose, including commercial applications, and to alter it
                and redistribute it freely, subject to the following
                restrictions:

                1. The origin of this software must not be misrepresented,
                    you must not claim that you wrote the original software.
                    If you use this software in a product, an acknowledgment
                    in the product documentation would be appreciated but is
                    not required.

                2. Altered source versions must be plainly marked as such, and
                   must not be misrepresented as being the original software.

                3. This notice may not be removed or altered from any source
                   distribution.

History:

Version 1.20
------------
20.06.2000      Bugfix for MimeEncodeStream: Wrong BUFFER_SIZE resulted in
                additional bytes stuffed inbetween the OutBuffer.

                Changed the order of the variables in the MimeEncode interface
                (moved InputBytesCount from last to 2nd, just after InputBuffer).
                Sorry for the inconvenience, but this way it is more consistent
                with MimeDecode. Now MimeEncode has the following variable order:

                procedure MimeEncode (
                  const InputBuffer: Pointer;
                  const InputBytesCount: Integer;
                  const OutputBuffer: Pointer);

                MimeDecode: Interface chage to make it a function: It now returns
                the number of bytes written to the output buffer not as a var
                but as the result of a function. This way coding is made
                simpler since not OutputCount variable needs to be defined.

                Introduced two new functions mostly for internal use:

                * MimeDecodePartial: This is necessary to decode large
                  blocks of data in multiple parts. On initialization,
                  call MimeDecodePartial with ByteBuffer := 0 and
                  ByteBufferSpace := 4. Then repeatedly call it again
                  with new data loaded into the buffer. At the end of all data,
                * MimeDecodePartialEnd writes the remaining bytes from ByteBuffer
                  to the outputbuffer (see MimeDecodeStream for an example).

                MimeDecodePartial_ are necessary to decode inconsitent data
                (with linebreaks, tabs, whitespace) in multiple parts
                like in MimeDecodeStream.

Version 1.10
------------
19.04.2000      Fixed a small bug in MimeEncode which sometimes screwed up
                the very first bytes of the encoded output.

                Added the following wrapper functions:
                * MimeEncodeString & MimeDecodeString
                * MimeEncodeStream & MimeDecodeStream

Version 1.01
------------
09.04.2000      Fixed a bug in MIME_DECODE_TABLE which caused wrong results
                decoding binary files.

Version 1.00
------------
17.01.2000      Initial Public Release

Copyright (c) 2000 Ralf Junker
}

interface

{$DEFINE RangChecking}
{ RangeChecking defines if an EMime exception will be raised if the
  InputBuffer or OutputBuffer are nil. If defined, unit SysUtils will be included. }

uses Classes
 {$IFDEF RangeChecking}
 , SysUtils
 {$ENDIF};

function MimeEncodeString (const s: AnsiString): AnsiString;
{ MimeEncodeString takes a string, encodes it, and returns the result as a string.
  To decode the result string, use MimeDecodeString. }

function MimeDecodeString (const s: AnsiString): AnsiString;
{ MimeDecodeString takes a a string, decodes it, and returns the result as a string.
  Use MimeDecodeString to decode a string previously encoded with MimeEncodeString. }

procedure MimeEncodeStream (const InputStream: TStream; const OutputStream: TStream);
{ MimeEncodeStream encodes InputStream starting at the current position
  up to the end and writes the result to OutputStream, again starting at
  the current position. When done, it will not reset either stream's positions,
  but leave InputStream at the last read position (i.e. the end) and
  OutputStream at the last write position (which can, but most not be the end).
  To encode the entire InputStream from beginning to end, make sure
  that its offset is positioned at the beginning of the stream. You can
  force this by issuing Seek (0, soFromBeginning) before calling this function. }

procedure MimeDecodeStream (const InputStream: TStream; const OutputStream: TStream);
{ MimeDecodeStream decodes InputStream starting at the current position
  up to the end and writes the result to OutputStream, again starting at
  the current position. When done, it will not reset either stream's positions,
  but leave InputStream at the last read position (i.e. the end) and
  OutputStream at the last write position (which can, but most not be the end).
  To decode the entire InputStream from beginning to end, make sure
  that its offset is positioned at the beginning of the stream. You can
  force this by issuing Seek (0, soFromBeginning) before calling this function. }

function MimeEncodedSize (const i: Integer): Integer;
{ Calculates the output size of i MimeEncoded bytes. Use for MimeEncode only. }

function MimeDecodedSize (const i: Integer): Integer;
{ Calculates the maximum output size of i MimeDecoded bytes.
  You may use it for MimeDecode to calculate the maximum amount of memory
  required for decoding in one single pass. }

procedure MimeEncode (const InputBuffer: Pointer; const InputByteCount: Integer; const OutputBuffer: Pointer);
{ The primary Mime encoding routine.

  CAUTTION: OutputBuffer must have enough memory allocated to take all encoded output.
  MimeEncodedSize (InputBytesCount) calculates this amount in bytes. MimeEncode will
  then fill the entire OutputBuffer, so there is no OutputBytesCount result for
  this procedure. Preallocating all memory at once (as required by MimeEncode)
  avoids the time-cosuming process of reallocation.

  If not all data fits into memory at once, you can use MimeEncode multiple times,
  but you must be very careful about the size of the InputBuffer.
  See comments on BUFFER_SIZE below for details. }

function MimeDecode (const InputBuffer: Pointer; const InputBytesCount: Integer; const OutputBuffer: Pointer): Integer;
{ The primary Mime decoding routines.

  CAUTTION: OutputBuffer must have enough memory allocated to take all output.
  MimeDecodedSize (InputBytesCount) calculates this amount in bytes. There is
  no guarantee that all output will be filled after decoding. All decoding
  functions therefore return the acutal number of bytes written to OutputBuffer.
  Preallocating all memory at once (as is required by MimeDecode)
  avoids the time-cosuming process of reallocation. After calling
  MimeDecode, simply cut the allocated memory down to OutputBytesCount,
  i.e. SetLength (OutString, OutputBytesCount).

  If not all data fits into memory at once, you may NOT use MimeDecode multiple times.
  Instead, you must use the MimeDecodePartial_ functions.
  See MimeDecodeStream for an example. }

function MimeDecodePartial (const InputBuffer: Pointer; const InputBytesCount: Integer; const OutputBuffer: Pointer; var ByteBuffer: Cardinal; var ByteBufferSpace: Cardinal): Integer;
function MimeDecodePartialEnd (const OutputBuffer: Pointer; const ByteBuffer: Cardinal; const ByteBufferSpace: Cardinal): Integer;
{ The MimeDecodePartial_ functions are mostly for internal use.
  They serve the purpose of decoding very large data in multiple parts of
  smaller chunks, as used in MimeDecodeStream. }

{$IFDEF RangeChecking}
resourcestring
 SInputBufferNil              = 'MIME (Base64) Conversion: Input Buffer must not be NIL.';
 SOutputBufferNil             = 'MIME (Base64) Conversion: Output Buffer must not be NIL.';
 {$ENDIF}
 
const
 BUFFER_SIZE                  = $3000;
 { Caution: For MimeEncodeStream and all other kinds of multi-buffered
   Mime encodings (i.e. Files etc.), BufferSize must be set to a multiple of 3.
   Even though the implementation of the Mime decoding routines below
   does not require a particular buffer size, they work fastest with sizes of
   multiples of four. The chosen size is a multiple of 3 and of 4 as well.
   The following numbers are, in addition, also divisible by 1024:
   $2400, $3000, $3C00, $4800, $5400, $6000, $6C00. }
 
 MIME_ENCODE_TABLE            : array[ 0..63] of Byte = (
  065, 066, 067, 068, 069, 070, 071, 072, // 00 - 07
  073, 074, 075, 076, 077, 078, 079, 080, // 08 - 15
  081, 082, 083, 084, 085, 086, 087, 088, // 16 - 23
  089, 090, 097, 098, 099, 100, 101, 102, // 24 - 31
  103, 104, 105, 106, 107, 108, 109, 110, // 32 - 39
  111, 112, 113, 114, 115, 116, 117, 118, // 40 - 47
  119, 120, 121, 122, 048, 049, 050, 051, // 48 - 55
  052, 053, 054, 055, 056, 057, 043, 047); // 56 - 63
 
 MIME_DECODE_TABLE            : array[ Byte] of Cardinal = (
  255, 255, 255, 255, 255, 255, 255, 255, //  00 -  07
  255, 255, 255, 255, 255, 255, 255, 255, //  08 -  15
  255, 255, 255, 255, 255, 255, 255, 255, //  16 -  23
  255, 255, 255, 255, 255, 255, 255, 255, //  24 -  31
  255, 255, 255, 255, 255, 255, 255, 255, //  32 -  39
  255, 255, 255, 062, 255, 255, 255, 063, //  40 -  47
  052, 053, 054, 055, 056, 057, 058, 059, //  48 -  55
  060, 061, 255, 255, 255, 255, 255, 255, //  56 -  63
  255, 000, 001, 002, 003, 004, 005, 006, //  64 -  71
  007, 008, 009, 010, 011, 012, 013, 014, //  72 -  79
  015, 016, 017, 018, 019, 020, 021, 022, //  80 -  87
  023, 024, 025, 255, 255, 255, 255, 255, //  88 -  95
  255, 026, 027, 028, 029, 030, 031, 032, //  96 - 103
  033, 034, 035, 036, 037, 038, 039, 040, // 104 - 111
  041, 042, 043, 044, 045, 046, 047, 048, // 112 - 119
  049, 050, 051, 255, 255, 255, 255, 255, // 120 - 127
  255, 255, 255, 255, 255, 255, 255, 255,
  255, 255, 255, 255, 255, 255, 255, 255,
  255, 255, 255, 255, 255, 255, 255, 255,
  255, 255, 255, 255, 255, 255, 255, 255,
  255, 255, 255, 255, 255, 255, 255, 255,
  255, 255, 255, 255, 255, 255, 255, 255,
  255, 255, 255, 255, 255, 255, 255, 255,
  255, 255, 255, 255, 255, 255, 255, 255,
  255, 255, 255, 255, 255, 255, 255, 255,
  255, 255, 255, 255, 255, 255, 255, 255,
  255, 255, 255, 255, 255, 255, 255, 255,
  255, 255, 255, 255, 255, 255, 255, 255,
  255, 255, 255, 255, 255, 255, 255, 255,
  255, 255, 255, 255, 255, 255, 255, 255,
  255, 255, 255, 255, 255, 255, 255, 255,
  255, 255, 255, 255, 255, 255, 255, 255);
 
type
 {$IFDEF RangeChecking}
 EMime = class (Exception);
 {$ENDIF}
 
 PByte4 = ^TByte4;
 TByte4 = packed record
  b1: Byte;
  b2: Byte;
  b3: Byte;
  b4: Byte;
 end;
 
 PByte3 = ^TByte3;
 TByte3 = packed record
  b1: Byte;
  b2: Byte;
  b3: Byte;
 end;
 
implementation

{ **************************************************************************** }
{ Wrapper functions & procedures
{ **************************************************************************** }

function MimeEncodeString (const s: AnsiString): AnsiString;
label
 NothingToDo;
var
 l                            : Integer;
begin
 if Pointer (s) = nil then goto NothingToDo;
 
 { l := length (s); }
 l := Integer (Pointer (Integer (Pointer (s)) - 4)^);
 
 SetString (Result, nil, MimeEncodedSize (l));
 if Pointer (Result) = nil then goto NothingToDo;
 
 MimeEncode (Pointer (s), l, Pointer (Result));
 Exit;
 
 NothingToDo:
 Result := '';
end;

{ ********** }

function MimeDecodeString (const s: AnsiString): AnsiString;
label
 NothingToDo;
var
 ByteBuffer, ByteBufferSpace  : Cardinal;
 l                            : Integer;
begin
 if Pointer (s) = nil then goto NothingToDo;
 
 { l := length (s); }
 l := Integer (Pointer (Integer (Pointer (s)) - 4)^);
 
 SetString (Result, nil, (l + 3) div 4 * 3);
 if Pointer (Result) = nil then goto NothingToDo;
 
 ByteBuffer := 0;
 ByteBufferSpace := 4;
 
 l := MimeDecodePartial (Pointer (s), l, Pointer (Result), ByteBuffer, ByteBufferSpace);
 Inc (l, MimeDecodePartialEnd (Pointer (Integer (Result) + l), ByteBuffer, ByteBufferSpace));
 
 SetLength (Result, l);
 Exit;
 
 NothingToDo:
 Result := '';
end;

{ **************************************************************************** }

procedure MimeEncodeStream (const InputStream: TStream; const OutputStream: TStream);
var
 InputBuffer                  : array[ 0..BUFFER_SIZE - 1] of Byte;
 OutputBuffer                 : array[ 0.. ((BUFFER_SIZE + 2) div 3) * 4 - 1] of Byte;
 BytesRead                    : Integer;
begin
 BytesRead := InputStream.Read (InputBuffer, SizeOf (InputBuffer));
 while BytesRead = SizeOf (InputBuffer) do
  begin
   MimeEncode (@InputBuffer, SizeOf (InputBuffer), @OutputBuffer);
   OutputStream.Write (OutputBuffer, SizeOf (OutputBuffer));
   BytesRead := InputStream.Read (InputBuffer, SizeOf (InputBuffer));
  end;
 if BytesRead > 0 then
  begin
   MimeEncode (@InputBuffer, BytesRead, @OutputBuffer);
   OutputStream.Write (OutputBuffer, MimeEncodedSize (BytesRead));
  end;
end;

{ ********** }

procedure MimeDecodeStream (const InputStream: TStream; const OutputStream: TStream);
var
 ByteBuffer, ByteBufferSpace  : Cardinal;
 InputBuffer                  : array[ 0..BUFFER_SIZE - 1] of Byte;
 OutputBuffer                 : array[ 0.. (BUFFER_SIZE + 3) div 4 * 3 - 1] of Byte;
 BytesRead                    : Integer;
begin
 ByteBuffer := 0;
 ByteBufferSpace := 4;
 BytesRead := InputStream.Read (InputBuffer, SizeOf (InputBuffer));
 while BytesRead > 0 do
  begin
   OutputStream.Write (OutputBuffer, MimeDecodePartial (@InputBuffer, BytesRead, @OutputBuffer, ByteBuffer, ByteBufferSpace));
   BytesRead := InputStream.Read (InputBuffer, SizeOf (InputBuffer));
  end;
 OutputStream.Write (OutputBuffer, MimeDecodePartialEnd (@OutputBuffer, ByteBuffer, ByteBufferSpace));
end;

{ **************************************************************************** }
{ Helper functions
{ **************************************************************************** }

function MimeEncodedSize (const i: Integer): Integer;
begin
 Result := (i + 2) div 3 * 4;
end;

function MimeDecodedSize (const i: Integer): Integer;
begin
 Result := (i + 3) div 4 * 3;
end;

{ **************************************************************************** }
{ Primary functions & procedures
{ **************************************************************************** }

procedure MimeEncode (const InputBuffer: Pointer; const InputByteCount: Integer; const OutputBuffer: Pointer);
var
 b                            : Cardinal;
 InMax3                       : Integer;
 pIn, PInLimit                : ^Byte;
 POut                         : PByte4;
begin
 {$IFDEF RangeChecking}
 if InputBuffer = nil then raise EMime.Create (SInputBufferNil);
 if OutputBuffer = nil then raise EMime.Create (SOutputBufferNil);
 {$ENDIF}
 if InputByteCount <= 0 then Exit;
 
 pIn := InputBuffer;
 InMax3 := InputByteCount div 3 * 3;
 POut := OutputBuffer;
 
 Integer (PInLimit) := Integer (pIn) + InMax3;
 
 while pIn <> PInLimit do
  begin
   b := pIn^;
   b := b shl 8;
   Inc (pIn);
   b := b or pIn^;
   b := b shl 8;
   Inc (pIn);
   b := b or pIn^;
   Inc (pIn);

   // Write 4 bytes to OutputBuffer (in reverse order).
   POut.b4 := MIME_ENCODE_TABLE[ b and $3F];
   b := b shr 6;
   POut.b3 := MIME_ENCODE_TABLE[ b and $3F];
   b := b shr 6;
   POut.b2 := MIME_ENCODE_TABLE[ b and $3F];
   b := b shr 6;
   POut.b1 := MIME_ENCODE_TABLE[ b];

   Inc (POut);
  end;
 
 case InputByteCount - InMax3 of
  1:
   begin
    b := pIn^;

    b := b shl 4;

    POut.b2 := MIME_ENCODE_TABLE[ b and $3F];
    b := b shr 6;
    POut.b1 := MIME_ENCODE_TABLE[ b];

    POut.b3 := Ord ('='); // Fill remaining 2 bytes.
    POut.b4 := Ord ('=');
   end;
  2:
   begin
    b := pIn^;
    Inc (pIn);
    b := b shl 8;
    b := b or pIn^;

    b := b shl 2;

    POut.b3 := MIME_ENCODE_TABLE[ b and $3F];
    b := b shr 6;
    POut.b2 := MIME_ENCODE_TABLE[ b and $3F];
    b := b shr 6;
    POut.b1 := MIME_ENCODE_TABLE[ b];

    POut.b4 := Ord ('='); // Fill remaining byte.
   end;
 end;
end;

{ ********** }

function MimeDecode (const InputBuffer: Pointer; const InputBytesCount: Integer; const OutputBuffer: Pointer): Integer;
var
 ByteBuffer, ByteBufferSpace  : Cardinal;
begin
 ByteBuffer := 0;
 ByteBufferSpace := 4;
 Result := MimeDecodePartial (InputBuffer, InputBytesCount, OutputBuffer, ByteBuffer, ByteBufferSpace);
 Inc (Result, MimeDecodePartialEnd (Pointer (Integer (OutputBuffer) + Result), ByteBuffer, ByteBufferSpace));
end;

{ ********** }

function MimeDecodePartial (const InputBuffer: Pointer; const InputBytesCount: Integer; const OutputBuffer: Pointer; var ByteBuffer: Cardinal; var ByteBufferSpace: Cardinal): Integer;
var
 lByteBuffer, lByteBufferSpace, c: Cardinal;
 pIn, PInLimit                : ^Byte;
 POut                         : PByte3;
begin
 {$IFDEF RangeChecking}
 if InputBuffer = nil then raise EMime.Create (SInputBufferNil);
 if OutputBuffer = nil then raise EMime.Create (SOutputBufferNil);
 {$ENDIF}
 if InputBytesCount > 0 then
  begin
   pIn := InputBuffer;
   Integer (PInLimit) := Integer (pIn) + InputBytesCount;

   POut := OutputBuffer;

   lByteBuffer := ByteBuffer;
   lByteBufferSpace := ByteBufferSpace;

   while pIn <> PInLimit do
    begin
     c := MIME_DECODE_TABLE[ pIn^]; // Read from InputBuffer.
     Inc (pIn);

     if c = $FF then Continue;

     lByteBuffer := lByteBuffer shl 6;
     lByteBuffer := lByteBuffer or c;
     Dec (lByteBufferSpace);

     if lByteBufferSpace <> 0 then Continue; // Read 4 bytes from InputBuffer?

     POut.b3 := Byte (lByteBuffer); // Write 3 bytes to OutputBuffer (in reverse order).
     lByteBuffer := lByteBuffer shr 8;
     POut.b2 := Byte (lByteBuffer);
     lByteBuffer := lByteBuffer shr 8;
     POut.b1 := Byte (lByteBuffer);
     lByteBuffer := 0;
     Inc (POut);

     lByteBufferSpace := 4;
    end;

   ByteBuffer := lByteBuffer;
   ByteBufferSpace := lByteBufferSpace;

   Result := Integer (POut) - Integer (OutputBuffer);
  end
 else
  Result := 0;
end;

{ ********** }

function MimeDecodePartialEnd (const OutputBuffer: Pointer; const ByteBuffer: Cardinal; const ByteBufferSpace: Cardinal): Integer;
var
 lByteBuffer                  : Cardinal;
begin
 {$IFDEF RangeChecking}
 if OutputBuffer = nil then raise EMime.Create (SOutputBufferNil);
 {$ENDIF}
 case ByteBufferSpace of
  1:
   begin
    lByteBuffer := ByteBuffer shr 2;
    PByte3 (OutputBuffer).b2 := Byte (lByteBuffer);
    lByteBuffer := lByteBuffer shr 8;
    PByte3 (OutputBuffer).b1 := Byte (lByteBuffer);
    Result := 2;
   end;
  2:
   begin
    lByteBuffer := ByteBuffer shr 4;
    PByte3 (OutputBuffer).b1 := Byte (lByteBuffer);
    Result := 1;
   end;
 else
  Result := 0;
 end;
end;

end.

